-- ------------------------------------------------------------------------------------------------
-- Handles Button-layout for a Group.
-- ------------------------------------------------------------------------------------------------

local function L(text) return text end  -- Localization stub

local layouts     = {}
pol_layoutClasses = {}

function Pol_MonitorLayoutFactory(layoutType, ...)
	if not pol_layoutClasses[layoutType] then
		error(L("Layout type '%s' not found"):format(layoutType or "nil"))
	end
	layout = pol_layoutClasses[layoutType](...)
	table.insert(layouts, layout)
	return layout
end


-- Base -------------------------------------------------------------------------------------------

local LayoutBase             = polOO.Class()
LayoutBase._allLayoutFrames  = {}
LayoutBase._freeLayoutFrames = {}

function LayoutBase:init(nameOrProfile, point, relFrame, relPoint, offsetX, offsetY, scale)
	self.frame = self:_CreateFrame(self)

	if type(nameOrProfile) ~= "table" then
		-- We're being created on the fly by the user
		local x, y = UIParent:GetCenter()
		self.profile = {
			name = nameOrProfile,
			point = point or "TOPLEFT",
			relFrame = (relFrame or UIParent):GetName(),
			relPoint = relPoint or "BOTTOMLEFT",
			offsetX = offsetX or x,
			offsetY = offsetY or y,
			scale = scale or 100,
			locked = false,
		}
	else
		-- We're being recreated from a stored profile
		self.profile = nameOrProfile
		self.profile.scale = self.profile.scale or 100
		
		if self.profile.locked then
			self:LockPosition()
		end
	end
	
	self.frame:SetPoint(self.profile.point, getglobal(self.profile.relFrame) or UIParent, self.profile.relPoint, self.profile.offsetX, self.profile.offsetY)
	self.frame:SetScale(self.profile.scale/100)
end

function LayoutBase:delete()
	self:_ReleaseFrame(self.frame)
	self.frame = nil
end

function LayoutBase:_CreateFrame(layout)
	local frame = nil
	
	-- See if we have one ready
	if #self._freeLayoutFrames > 0 then
		frame = self._freeLayoutFrames[#self._freeLayoutFrames]
		self._freeLayoutFrames[#self._freeLayoutFrames] = nil
	else
		frame = CreateFrame("Button", "Pol_MonitorLayout"..#self._allLayoutFrames, UIParent)
		frame.dragFrame = CreateFrame("Button", nil, frame)
		table.insert(self._allLayoutFrames, frame)
	end

	frame.layout = layout

	frame:SetFrameLevel(0)
	frame:SetClampedToScreen(true)
	frame:EnableMouse(false)
	frame:SetMovable(true)
	frame:SetScale(1)

	frame.dragFrame:SetFrameLevel(1)
	frame.dragFrame:SetBackdrop({bgFile="Interface\\Tooltips\\UI-Tooltip-Background"});
    frame.dragFrame:SetBackdropColor(0, 0, 0, .5);
	frame.dragFrame:RegisterForDrag("LeftButton")
	frame.dragFrame:EnableMouse()

	frame.dragFrame:SetScript("OnDragStart", function() layout:OnDragStart() end)
	frame.dragFrame:SetScript("OnDragStop", function() layout:OnDragStop() end)
	frame.dragFrame:SetScript("OnEnter", function() layout:OnEnter() end)
	frame.dragFrame:SetScript("OnLeave", function() layout:OnLeave() end)

	frame:Show()
	return frame
end

function LayoutBase:_ReleaseFrame(frame)
	table.insert(self._freeLayoutFrames, frame)
	frame:Hide()
end

function LayoutBase:GetProfileTable()
	return self.profile
end

function LayoutBase:GetOptionsTable()
	error("Subclasses must override")
end

function LayoutBase:GetDescription()
	error("Subclasses must override")
end

function LayoutBase:GetUsage()
	error("Subclasses must override")
end

function LayoutBase:AddButton(button)
	error("Subclasses must override")
end

function LayoutBase:RemoveButton(button)
	error("Subclasses must override")
end

function LayoutBase:GetScale()
	return self.profile.scale
end

function LayoutBase:SetScale(scale)
	self.profile.scale = scale or 100
	
	-- Make the frame stand still while scaling (I peeked at how Bongos2 does this.. :)
	local point, parent, relpoint, x, y = self.frame:GetPoint()
	local newscale = self.profile.scale/100
	local ratio = self.frame:GetScale()/newscale
	x, y = x*ratio, y*ratio
	
	self.frame:SetScale(newscale)
	self.frame:SetPoint(point, parent, relpoint, x, y)
	
	self.profile.offsetX = x
	self.profile.offsetY = y
end

function LayoutBase:LockPosition()
	self.profile.locked = true;
	self.frame.dragFrame:Hide()
end

function LayoutBase:UnlockPosition()
	self.profile.locked = false;
	self.frame.dragFrame:Show()
end

function LayoutBase:IsPositionLocked()
	return self.frame.dragFrame:IsShown()
end

function LayoutBase:OnEnter()
	-- Show tooltip
	if self.frame.dragFrame:GetCenter() < UIParent:GetCenter() then
		GameTooltip:SetOwner(self.frame.dragFrame, "ANCHOR_RIGHT");
	else
		GameTooltip:SetOwner(self.frame.dragFrame, "ANCHOR_LEFT");
	end

    GameTooltip:ClearLines();
	GameTooltip:AddLine(L("Monitor group: %s"):format(self.profile.name), 1, 1, 1);
	GameTooltip:AddLine(L("Click and drag to move the group"), 1, .75, 0, 1);
	GameTooltip:Show();
end

function LayoutBase:OnLeave()
	GameTooltip:Hide();
end

function LayoutBase:OnDragStart()
	self.frame:StartMoving()
end

function LayoutBase:OnDragStop()
	self.frame:StopMovingOrSizing()
	self.profile.offsetX = self.frame:GetLeft()
	self.profile.offsetY = self.frame:GetTop()
end


-- LayoutGrid -------------------------------------------------------------------------------------

local LayoutGrid = polOO.Class(LayoutBase)
pol_layoutClasses.Grid = LayoutGrid

function LayoutGrid:init(nameOrProfile, point, relFrame, relPoint, offsetX, offsetY, rows, cols, rowsep, colsep, growLeft, growUp, scale)
	LayoutBase.init(self, nameOrProfile, point, relFrame, relPoint, offsetX, offsetY, scale)
	if type(nameOrProfile) ~= "table" then
		self.profile.rows = rows
		self.profile.cols = cols
		self.profile.rowsep = rowsep or 4
		self.profile.colsep = colsep or 4
		self.profile.growLeft = growLeft
		self.profile.growUp = growUp
		self.profile.scale = scale or 100
	end
	self.buttons = {}
	self:Layout()
end

function LayoutGrid:delete()
	-- The Group will take care of the buttons, we just clean up our reference to them.
	for k, button in pairs(self.buttons) do
		button:GetFrame():SetParent(UIParent)
		self.buttons[k] = nil
	end
	LayoutBase.delete(self)
end

function LayoutGrid:GetDescription()
	return L("Basic grid layout")
end

function LayoutGrid:GetUsage()
	return L("[point] [relativeFrame] [relativePoint] [offsetX] [offsetY] [rows] [cols] [rowSeparation] [colSeparation] [growLeft] [growUp]")
end

function LayoutGrid:GetOptionsTable(order, startOrder)
	order, startOrder = order or 100, startOrder or 1
	
	if not self.optionsTable then
		self.optionsTable = {
			type = "group",
			name = L("Layout"),
			desc = L("Manage the layout"),
			handler = self,
			order = order,
			args = {
				rows = {
					type = "input",
					name = L("Rows"),
					desc = L("Number of rows.\nInfinite if blank"),
					get = function(info) return self:GetRows() end,
					set = function(info, v) self:SetRows(v) end,
					usage = "<number or rows>",
					order = startOrder + 0,
				},
				cols = {
					type = "input",
					name = L("Columns"),
					desc = L("Number of column.\nInfinite if blank"),
					get = function(info) return self:GetCols() end,
					set = function(info, v) self:SetCols(v) end,
					usage = "<number or columns>",
					order = startOrder + 1,
				},
				rowsep = {
					type = "input",
					name = L("Row separation"),
					desc = L("Distance between rows"),
					get = function(info) return self:GetRowSeparation() end,
					set = function(info, v) self:SetRowSeparation(v) end,
					usage = "<row separation>",
					order = startOrder + 2,
				},
				colsep = {
					type = "input",
					name = L("Column separation"),
					desc = L("Distance between columns"),
					get = function(info) return self:GetColSeparation() end,
					set = function(info, v) self:SetColSeparation(v) end,
					usage = "<column separation>",
					order = startOrder + 3,
				},
				growleft = {
					type = "toggle",
					name = L("Grow left"),
					desc = L("If set, rows grow to the left"),
					get = function(info) return self:GetGrowLeft() end,
					set = function(info, v) self:SetGrowLeft(v) end,
					order = startOrder + 4,
				},
				growup = {
					type = "toggle",
					name = L("Grow up"),
					desc = L("If set, columns grow upwards"),
					get = function(info) return self:GetGrowUp() end,
					set = function(info, v) self:SetGrowUp(v) end,
					order = startOrder + 5,
				},
				scale = {
					type = "range",
					name = L("Scale"),
					desc = L("UI scale for this group"),
					min = 50,
					max = 500,
					step = 1,
					get = function(info) return self:GetScale() end,
					set = function(info, v) self:SetScale(v) end,
					order = startOrder + 6,
				},
			},
		}
	end
	
	return self.optionsTable
end

function LayoutGrid:GetRows()
	return tostring(self.profile.rows or "")
end

function LayoutGrid:SetRows(rows)
	self.profile.rows = tonumber(rows)
	self:Layout()
end

function LayoutGrid:GetCols()
	return tostring(self.profile.cols or "")
end

function LayoutGrid:SetCols(cols)
	self.profile.cols = tonumber(cols)
	self:Layout()
end

function LayoutGrid:GetRowSeparation()
	return tostring(self.profile.rowsep or "")
end

function LayoutGrid:SetRowSeparation(rowsep)
	self.profile.rowsep = tonumber(rowsep)
	self:Layout()
end

function LayoutGrid:GetColSeparation()
	return tostring(self.profile.colsep or "")
end

function LayoutGrid:SetColSeparation(colsep)
	self.profile.colsep = tonumber(colsep)
	self:Layout()
end

function LayoutGrid:GetGrowLeft()
	return self.profile.growLeft
end

function LayoutGrid:SetGrowLeft(growLeft)
	self.profile.growLeft = growLeft
	self:Layout()
end

function LayoutGrid:GetGrowUp()
	return self.profile.growUp
end

function LayoutGrid:SetGrowUp(growUp)
	self.profile.growUp = growUp
	self:Layout()
end

function LayoutGrid:MoveButton(fromPos, toPos)
	assert(fromPos > 0 and fromPos <= #self.buttons, "Trying to move a button from an invalid position")
	assert(toPos > 0 and toPos <= #self.buttons, "Trying to move a button to an invalid position")
	local temp = self.buttons[fromPos]
	self.buttons[fromPos] = self.buttons[toPos]
	self.buttons[toPos] = temp
	self:Layout()
end

function LayoutGrid:AddButton(button)
	table.insert(self.buttons, button)
	button:GetFrame():SetParent(self.frame)
	button:GetFrame():SetFrameLevel(2)
	self:Layout()
end

function LayoutGrid:RemoveButton(button)
	for k, v in pairs(self.buttons) do
		if v == button then
			button:GetFrame():SetParent(UIParent)
			table.remove(self.buttons, k)
			break
		end
	end
	self:Layout()
end

function LayoutGrid:Layout()
	local growUp, growLeft = self.profile.growUp, self.profile.growLeft
	local thepoint = (growUp and "BOTTOM" or "TOP")..(growLeft and "RIGHT" or "LEFT")
	local buttonWidth, buttonHeight = 36, 36

	-- Resize our frame to match the first button to help the dragFrame align itself properly
	if self.buttons[1] then
		buttonWidth = self.buttons[1]:GetWidth()
		buttonHeight = self.buttons[1]:GetHeight()
	end
	
	self.frame:SetWidth(buttonWidth)
	self.frame:SetHeight(buttonHeight)

	-- Re-position all buttons according to our current configuration
	local rowCount, colCount = 0, 0
	
	for i, button in pairs(self.buttons) do
		local point, relFrame, relPoint = self.profile.point, self.frame, self.profile.point
		local offsetX, offsetY = 0, 0
		local index = i - 1
		local row, col
		
		if self.profile.cols or not self.profile.rows then
			-- Grow column, then row
			local cols = self.profile.cols or (#self.buttons)
			row = floor(index/cols)
			col = index%cols
			
			if col == 0 then
				-- Bind to col 0 in previous row, or layout frame if first row
				if row > 0 then
					relFrame = self.buttons[cols*(row-1)+1]:GetFrame()
					point = growUp and "BOTTOM" or "TOP"
					relPoint = growUp and "TOP" or "BOTTOM"
					offsetY = growUp and self.profile.rowsep or -self.profile.rowsep
					offsetX = 0
				end
			else
				-- Bind to previous col in this row
				relFrame = self.buttons[cols*(row)+col]:GetFrame()
				point = growLeft and "RIGHT" or "LEFT"
				relPoint = growLeft and "LEFT" or "RIGHT"
				offsetX = growLeft and -self.profile.colsep or self.profile.colsep
				offsetY = 0
			end
		else
			-- Grow row, then column
			row = index%self.profile.rows
			col = floor(index/self.profile.rows)
			
			if col == 0 then
				-- Bind to col 0 in previous row, or layout frame if first row
				if row > 0 then
					relFrame = self.buttons[row]:GetFrame()
					point = growUp and "BOTTOM" or "TOP"
					relPoint = growUp and "TOP" or "BOTTOM"
					offsetY = growUp and self.profile.rowsep or -self.profile.rowsep
					offsetX = 0
				end
			else
				-- Bind to previous col in this row
				relFrame = self.buttons[self.profile.rows*(col-1)+row+1]:GetFrame()
				point = growLeft and "RIGHT" or "LEFT"
				relPoint = growLeft and "LEFT" or "RIGHT"
				offsetX = growLeft and -self.profile.colsep or self.profile.colsep
				offsetY = 0
			end
		end
		
		button:ClearAllPoints()
		button:SetPoint(point, relFrame, relPoint, offsetX, offsetY)

		if rowCount < row then rowCount = row end
		if colCount < col then colCount = col end
	end
	
	-- Fit drag frame to buttons
	self.frame.dragFrame:ClearAllPoints()
	self.frame.dragFrame:SetPoint(thepoint, self.frame, thepoint, growLeft and 6 or -6, growUp and -6 or 6)
	self.frame.dragFrame:SetWidth((buttonWidth + self.profile.colsep)*(colCount + 1) - self.profile.colsep + 12)
	self.frame.dragFrame:SetHeight((buttonHeight + self.profile.rowsep)*(rowCount + 1) - self.profile.rowsep + 12)
end

