-- ------------------------------------------------------------------------------------------------
-- Defines a base class for classes that need to handle events
-- ------------------------------------------------------------------------------------------------

-- Short-circuit if we've already been loaded
if PolEvents then return end

PolEvents = polOO.Class()
local eventFrame = CreateFrame("Frame", "PolEvents_EventFrame", UIParent)
local onUpdateTable = {}
local eventTable = {}


function PolEvents:delete()
	self:UnRegisterAllEvents()
	self:StopOnUpdate()
end

function PolEvents:RegisterEvent(event)
	if type(self[event]) ~= "function" then
		error("No handler defined for event '"..event.."'.")
	end

	if not next(eventTable) then
		eventFrame:SetScript("OnEvent",
			function(_, event, ...)
				if eventTable[event] then
					for obj, v in pairs(eventTable[event]) do
						obj[event](obj, ...)
					end
				end
			end)
	end
	
	if not eventTable[event] then
		eventTable[event] = {}
		eventFrame:RegisterEvent(event)
	end
	
	eventTable[event][self] = true
end

function PolEvents:UnRegisterEvent(event)
	if eventTable[event] then
		eventTable[event][self] = nil
	end
	
	if not next(eventTable[event]) then
		eventTable[event] = nil
		
		if not next(eventTable) then
			eventFrame:UnregisterEvent(event)
			eventTable[event] = nil
		end
	end
end

function PolEvents:UnRegisterAllEvents()
	for event, d in pairs(eventTable) do
		self:UnRegisterEvent(event)
	end
end

function PolEvents:StartOnUpdate(period)
	if self:IsOnUpdateRunning() then
		self._period = period or 0
		self._nextUpdate = GetTime() + self._period
	else
		self._period = period or 0
		self._nextUpdate = 1
		
		if not next(onUpdateTable) then
			eventFrame:SetScript("OnUpdate",
				function(_, ...)
					for obj, v in pairs(onUpdateTable) do
						obj._OnUpdate(obj, ...)
					end
				end)
		end
		
		onUpdateTable[self] = true
	end
end

function PolEvents:StopOnUpdate()
	if self:IsOnUpdateRunning() then
		onUpdateTable[self] = nil
		if not next(onUpdateTable) then
			eventFrame:SetScript("OnUpdate", nil)
		end
		self._nextUpdate = nil
	end
end

function PolEvents:IsOnUpdateRunning()
	return onUpdateTable[self] ~= nil
end

function PolEvents:OnUpdate(...)
	error("Subclasses must override")
end

function PolEvents:_OnUpdate(...)
	if self._nextUpdate and self._nextUpdate <= GetTime() then
		self._nextUpdate = GetTime() + self._period
		self:OnUpdate(...)
	end
end

